@@ -34,7 +34,6 @@
  *
  *
  */
-/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/s3/s3_driver.c,v 1.19tsi Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -42,10 +41,8 @@
 
 #include "xf86.h"
 #include "xf86_OSproc.h"
-#include "xf86_ansic.h"
 #include "xf86Pci.h"
 #include "xf86PciInfo.h"
-#include "xf86Version.h"
 #include "xf86Resources.h"
 #include "xf86fbman.h"
 #include "xf86cmap.h"
@@ -56,6 +53,7 @@
 #include "micmap.h"
 #include "mibstore.h"
 #include "fb.h"
+#include "inputstr.h"
 
 #include "IBM.h"
 #include "TI.h"
@@ -154,20 +152,22 @@ static PciChipsets S3PciChipsets[] = {
 
 typedef enum {
 	OPTION_NOACCEL,
-	OPTION_SWCURS,
+	OPTION_HWCURS,
 	OPTION_SLOW_DRAM_REFRESH,
 	OPTION_SLOW_DRAM,
 	OPTION_SLOW_EDODRAM,
-	OPTION_SLOW_VRAM
+	OPTION_SLOW_VRAM,
+	OPTION_XVIDEO
 } S3Opts;
 
 static OptionInfoRec S3Options[] = {
 	{ OPTION_NOACCEL, "noaccel", OPTV_BOOLEAN, {0}, FALSE },
-	{ OPTION_SWCURS, "swcursor", OPTV_BOOLEAN, {0}, FALSE },
+	{ OPTION_HWCURS, "hwcursor", OPTV_BOOLEAN, {0}, FALSE },
 	{ OPTION_SLOW_DRAM_REFRESH, "slow_dram_refresh", OPTV_BOOLEAN, {0}, FALSE },
 	{ OPTION_SLOW_DRAM, "slow_dram", OPTV_BOOLEAN, {0}, FALSE },
 	{ OPTION_SLOW_EDODRAM, "slow_edodram", OPTV_BOOLEAN, {0}, FALSE },
 	{ OPTION_SLOW_VRAM, "slow_vram", OPTV_BOOLEAN, {0}, FALSE },
+	{ OPTION_XVIDEO, "XVideo", OPTV_BOOLEAN, {0}, FALSE },
 	{ -1, NULL, OPTV_NONE, {0}, FALSE }
 };
 
@@ -390,21 +390,20 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
         pScrn->monitor = pScrn->confScreen->monitor;
         
         if (!xf86SetDepthBpp(pScrn, 0, 0, 0, Support24bppFb | Support32bppFb))
-                return FALSE;
+	    return FALSE;
 
         switch (pScrn->depth) {
-            	case 8:
-                case 15:
-                case 16:
-                case 24:
-                case 32:
-      	               /* OK */
-                       break;
-                default:
-                       xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
-                                  "Given depth (%d) is not  supported by this driver\n",
-                                   pScrn->depth);
-	                       return FALSE;
+	case 8:
+	case 15:
+	case 16:
+	case 24:
+		/* OK */
+		break;
+	default:
+		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+			   "Given depth (%d) is not  supported by this driver\n",
+			   pScrn->depth);
+		return FALSE;
         }
 
         xf86PrintDepthBpp(pScrn);
@@ -429,37 +428,13 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
         xf86CollectOptions(pScrn, NULL);
         xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, S3Options);
          
-        if (xf86ReturnOptValBool(S3Options, OPTION_NOACCEL, FALSE)) {
-                pS3->NoAccel = TRUE;
-                xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: NoAccel - acceleration disabled\n");
-        } else          
-                pS3->NoAccel = FALSE;
-	if (xf86ReturnOptValBool(S3Options, OPTION_SWCURS, FALSE)) {
-		pS3->SWCursor = TRUE;
-		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: SWCursor - using software cursor\n");
-        } else
-		pS3->SWCursor = FALSE;
-	if (xf86ReturnOptValBool(S3Options, OPTION_SLOW_DRAM_REFRESH, FALSE)) {
-		pS3->SlowDRAMRefresh = TRUE;
-		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: Slow DRAM Refresh enabled\n");
-	} else
-		pS3->SlowDRAMRefresh = FALSE;
-	if (xf86ReturnOptValBool(S3Options, OPTION_SLOW_DRAM, FALSE)) {
-		pS3->SlowDRAM = TRUE;
-		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: Slow DRAM enabled\n");
-	} else
-		pS3->SlowDRAM = FALSE;
-	if (xf86ReturnOptValBool(S3Options, OPTION_SLOW_EDODRAM, FALSE)) {
-		pS3->SlowEDODRAM = TRUE;
-		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: Slow EDO DRAM enabled\n");
-	} else
-		pS3->SlowEDODRAM = FALSE;
-	if (xf86ReturnOptValBool(S3Options, OPTION_SLOW_DRAM, FALSE)) {
-		pS3->SlowVRAM = TRUE;
-		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Option: Slow VRAM enabled\n");
-	} else
-		pS3->SlowVRAM = FALSE;
-
+	pS3->XVideo = xf86ReturnOptValBool(S3Options, OPTION_XVIDEO, TRUE);
+	pS3->NoAccel = xf86ReturnOptValBool(S3Options, OPTION_NOACCEL, FALSE);
+	pS3->HWCursor = xf86ReturnOptValBool(S3Options, OPTION_HWCURS, FALSE);
+	pS3->SlowDRAMRefresh = xf86ReturnOptValBool(S3Options, OPTION_SLOW_DRAM_REFRESH, FALSE);
+	pS3->SlowDRAM = xf86ReturnOptValBool(S3Options, OPTION_SLOW_DRAM, FALSE);
+	pS3->SlowEDODRAM = xf86ReturnOptValBool(S3Options, OPTION_SLOW_EDODRAM, FALSE);
+	pS3->SlowVRAM = xf86ReturnOptValBool(S3Options, OPTION_SLOW_VRAM, FALSE);
         if (pScrn->numEntities > 1) {      
                 S3FreeRec(pScrn);
                 return FALSE;
@@ -500,7 +475,7 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
                 xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "ChipID override: 0x%04X\n",
                            pS3->Chipset);
         } else {
-                pS3->Chipset = pS3->PciInfo->chipType;   
+  	        pS3->Chipset = PCI_DEV_DEVICE_ID(pS3->PciInfo);
                 pScrn->chipset = (char *)xf86TokenToString(S3Chipsets,
                                                            pS3->Chipset);
         }                                                  
@@ -509,14 +484,16 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
                 xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "ChipRev override: %d\n",
                            pS3->ChipRev);
         } else
-                pS3->ChipRev = pS3->PciInfo->chipRev;    
+	        pS3->ChipRev = PCI_DEV_REVISION(pS3->PciInfo);
         
         xfree(pEnt);
         
         xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Chipset: \"%s\"\n", pScrn->chipset);
         
+#ifndef XSERVER_LIBPCIACCESS
         pS3->PciTag = pciTag(pS3->PciInfo->bus, pS3->PciInfo->device,
                              pS3->PciInfo->func);        
+#endif
 
 	switch (pS3->Chipset) {
 	case PCI_CHIP_964_0:
@@ -532,7 +509,12 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
 		break;
 	}
 
-	pS3->FBAddress = pS3->PciInfo->memBase[0];
+	if (HAS_STREAMS_PROCESSOR() && pS3->S3NewMMIO)
+        	pS3->hasStreams = TRUE;
+        else
+        	pS3->hasStreams = FALSE;
+ 
+	pS3->FBAddress = PCI_REGION_BASE(pS3->PciInfo, 0, REGION_MEM);
 	pScrn->memPhysBase = pS3->FBAddress;
 	pScrn->fbOffset = 0;
 	
@@ -581,9 +563,10 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
 
 	if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)
 	{
+          /* disable DAC power saving to avoid bright left edge */
 	  outb (0x3d4, 0x86);
 	  outb (0x3d5, 0x80);
-	  
+	  /* disable the stream display fetch length control */
 	  outb (0x3d4, 0x90);
 	  outb (0x3d5, 0x00);
 	}
@@ -664,17 +647,28 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
 #endif
 		switch(pScrn->bitsPerPixel) {
 		case 8:
-			pS3->MaxClock = 135000;
+			if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)
+				pS3->MaxClock = 170000;
+			else
+				pS3->MaxClock = 135000;
+
+			pScrn->rgbBits = 6;
 			break;
 		case 16:
 			pS3->MaxClock = 80000;
+			pScrn->rgbBits = 6;
 			break;
 		case 24:
 		case 32:
-			pS3->MaxClock = 50000;
+			if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)
+				pS3->MaxClock = 56700;
+			else
+				pS3->MaxClock = 50000;
+
+			pScrn->rgbBits = 8;
 			break;
 		}
-	        pScrn->rgbBits = 6;
+
 		pS3->LoadPalette = S3GenericLoadPalette;
 	}
 
@@ -684,7 +678,7 @@ static Bool S3PreInit(ScrnInfoPtr pScrn, int flags)
 		return FALSE;
 	}
 
-	if (pS3->SWCursor)
+	if (!pS3->HWCursor)
 		pS3->CursorInit = NULL;
 
 	pS3->RefClock = S3GetRefClock(pScrn);
@@ -749,6 +743,7 @@ static Bool S3ScreenInit(int scrnIndex, ScreenPtr pScreen, int argc,
 {
 	ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
 	S3Ptr pS3 = S3PTR(pScrn);
+	BoxRec ScreenArea;
 
 	pScrn->fbOffset = 0;
 
@@ -811,31 +806,71 @@ static Bool S3ScreenInit(int scrnIndex, ScreenPtr pScreen, int argc,
         miInitializeBackingStore(pScreen);
         xf86SetBackingStore(pScreen);
 
-	if (!pS3->NoAccel) {
-		if (pS3->S3NewMMIO) {
-			if (S3AccelInitNewMMIO(pScreen)) {
-				xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration enabled\n");
-				xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Using NewMMIO\n");
-			} else {
-				xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Acceleration initialization failed\n");
-				xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration disabled\n");
-			}
-		} else {
-			if (S3AccelInitPIO(pScreen)) {
-				xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration enabled\n");
-				xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Using PIO\n");
-			} else {
-				xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Acceleration initialization failed\n");
-				xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration disabled\n");
-			}
-		}
-	} else {
-		xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration disabled by option\n");
+	/* framebuffer manager setup */
+	ScreenArea.x1 = 0;
+	ScreenArea.y1 = 0;
+	ScreenArea.x2 = pScrn->displayWidth;
+	ScreenArea.y2 = (pScrn->videoRam * 1024) / pS3->s3BppDisplayWidth;
+
+	if (!xf86InitFBManager(pScreen, &ScreenArea)) {
+		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+			   "Memory manager initialization to (%d,%d) (%d,%d) failed\n",
+			   ScreenArea.x1, ScreenArea.y1,
+			   ScreenArea.x2, ScreenArea.y2);
+		return FALSE;
+	} else
+		xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+			   "Memory manager initialized to (%d,%d) (%d,%d)\n",
+			   ScreenArea.x1, ScreenArea.y1, 
+			   ScreenArea.x2, ScreenArea.y2);
+	
+
+	/* 2D acceleration setup */
+
+	if (pS3->NoAccel)
+		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
+			   "Acceleration disabled (by option)\n");
+
+	/* It seems that acceleration isn't supported for 24-bit packed
+	   colour. Disable it for S3 Trio64V2 */
+	if (!pS3->NoAccel && (pScrn->bitsPerPixel == 24) &&
+	    (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)) {
+		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Acceleration isn't supported for 24 bpp. Disabled.\n");
+		pS3->NoAccel = TRUE;
 	}
 
+	if (!pS3->NoAccel) {
+		if (pS3->S3NewMMIO)
+			if (S3AccelInitNewMMIO(pScreen)) {
+				xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+					   "Acceleration enabled\n");
+				xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+					   "Using NewMMIO\n");
+			} else {
+				xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
+					   "Acceleration initialization failed\n");
+				xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+					   "Acceleration disabled\n");
+			}
+		else {
+			if (S3AccelInitPIO(pScreen)) {
+				xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+					   "Acceleration enabled\n");
+				xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+					   "Using PIO\n");
+			} else {
+				xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
+					   "Acceleration initialization failed\n");
+				xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+					   "Acceleration disabled\n");
+			}
+		}
+	}
+	
         miDCInitialize(pScreen, xf86GetPointerScreenFuncs());
 
-	/* hw cursor setup */
+	/* HW cursor setup */
+
 	if (pS3->CursorInit) {
 		if (pS3->CursorInit(pScreen))
 			xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Using HW cursor\n");
@@ -869,10 +904,54 @@ static Bool S3ScreenInit(int scrnIndex, ScreenPtr pScreen, int argc,
 	pScrn->racIoFlags = pScrn->racMemFlags = RAC_COLORMAP
 	    | RAC_FB | RAC_VIEWPORT | RAC_CURSOR;
 
-#if 0
-	S3InitVideo(pScreen);
-#endif
+	if (pS3->SlowEDODRAM)
+		xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+			   "SlowEDODRAM: Setting 2-cycle EDO\n");
+
+	if (pS3->SlowVRAM)
+		xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+			   "SlowVRAM: -RAS low time is 4.5 MCLKs\n");
+
+	if (pS3->SlowDRAM)
+		xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+			   "SlowDRAM: -RAS precharge time is 3.5 MCLKs\n");
+
+	if (pS3->SlowDRAMRefresh)
+		xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
+			   "SlowDRAMRefresh: three refresh cycles per scanline\n");
+
+	/* XVideo setup */
+
+	if (pS3->XVideo) {
+		if (!pS3->hasStreams) {
+			xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Overlay video isn't supported by video hardware. Disabled.\n");
+			pS3->XVideo = FALSE;
+		} else if (pScrn->bitsPerPixel < 16) {
+			xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Overlay video isn't supported for %d bpp. Disabled.\n", pScrn->bitsPerPixel);
+			pS3->XVideo = FALSE;
+		}
+	} else
+		xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
+			   "Overlay video disabled by option\n");
 	
+	/* At present time we support XV only for chips with New MMIO */
+	if ((pS3->XVideo) && (pS3->S3NewMMIO))
+		S3InitVideo(pScreen);
+	
+	switch (pScrn->bitsPerPixel) {
+	case 8:
+		pS3->Streams_FIFO = FIFO_PS16_SS8;
+		break;
+	case 15:
+	case 16:
+		pS3->Streams_FIFO = FIFO_PS12_SS12;
+		break;
+	case 24:
+	case 32: 
+		pS3->Streams_FIFO = FIFO_PS8_SS16;
+		break;
+	}
+		
         return TRUE;
 }                 
 
@@ -1017,9 +1096,25 @@ static Bool S3MapMem(ScrnInfoPtr pScrn)
 	S3Ptr pS3 = S3PTR(pScrn);
 
 	if (pS3->S3NewMMIO) {
+
+
+#ifndef XSERVER_LIBPCIACCESS
 		pS3->MMIOBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_MMIO,
 					      pS3->PciTag, pS3->IOAddress,
 					      S3_NEWMMIO_REGSIZE);
+#else
+		{
+			void** result = (void**)&pS3->MMIOBase;
+			int err = pci_device_map_range(pS3->PciInfo,
+						       pS3->IOAddress,
+						       S3_NEWMMIO_REGSIZE,
+						       PCI_DEV_MAP_FLAG_WRITABLE,
+						       result);
+			
+			if (err) 
+				return FALSE;
+		}
+#endif
 		if (!pS3->MMIOBase) {
 			xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 				   "Could not map MMIO\n");
@@ -1027,9 +1122,25 @@ static Bool S3MapMem(ScrnInfoPtr pScrn)
 		}
 	}
 
+#ifndef XSERVER_LIBPCIACCESS
 	pS3->FBBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_FRAMEBUFFER,
 				    pS3->PciTag, pS3->FBAddress,
 				    pScrn->videoRam * 1024);
+
+#else
+	{
+		void** result = (void**)&pS3->FBBase;
+		int err = pci_device_map_range(pS3->PciInfo,
+					       pS3->FBAddress,
+					       pScrn->videoRam * 1024,
+					       PCI_DEV_MAP_FLAG_WRITABLE |
+					       PCI_DEV_MAP_FLAG_WRITE_COMBINE,
+					       result);
+		
+		if (err) 
+			return FALSE;
+	}
+#endif
 	if (!pS3->FBBase) {
 		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 			   "Could not map framebuffer\n");
@@ -1046,11 +1157,21 @@ static void S3UnmapMem(ScrnInfoPtr pScrn)
 {
 	S3Ptr pS3 = S3PTR(pScrn);
 
-	if (pS3->S3NewMMIO)
+	if (pS3->S3NewMMIO) {
+#ifndef XSERVER_LIBPCIACCESS
 		xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pS3->MMIOBase,
 				S3_NEWMMIO_REGSIZE);
+#else
+		pci_device_unmap_range(pS3->PciInfo, pS3->MMIOBase, S3_NEWMMIO_REGSIZE);
+#endif
+	}
+	
+#ifndef XSERVER_LIBPCIACCESS
 	xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pS3->FBBase,
 			pScrn->videoRam * 1024);
+#else
+	pci_device_unmap_range(pS3->PciInfo, pS3->FBBase, pScrn->videoRam * 1024);
+#endif
 
 	return;
 }
@@ -1065,8 +1186,12 @@ static int S3GetPixMuxShift(ScrnInfoPtr pScrn)
 		shift = 1;	/* XXX IBMRGB */
 	else if (pS3->Chipset == PCI_CHIP_TRIO || 
 	         pS3->Chipset == PCI_CHIP_TRIO64UVP || 
-	         pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)
-		shift = -(pS3->s3Bpp >> 1);
+	         pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX) {
+		if (pS3->s3Bpp == 2)
+			shift = -1; 
+		else 
+			shift = 0;
+	}
 
 	return shift;
 }
@@ -1092,13 +1217,22 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	pS3->s3ScissB = ((pScrn->videoRam * 1024) / pS3->s3BppDisplayWidth) - 1;
 	pS3->s3ScissR = pScrn->displayWidth - 1;
 
-	if (mode->HTotal == mode->CrtcHTotal) {
+	/* 
+	   Set correct blanking for S3 Trio64V2. It's also needed
+	   to clear cr33_5.
+	*/
+	if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)
+		mode->CrtcHBlankStart = mode->CrtcHDisplay + 8;
+
+	if ((mode->HTotal == mode->CrtcHTotal) && (pS3->pixMuxShift != 0)) {
 		if (pS3->pixMuxShift > 0) {
 			/* XXX hack */
 /* 			mode->Flags |= V_PIXMUX; */
 
 			mode->CrtcHTotal >>= pS3->pixMuxShift;
 			mode->CrtcHDisplay >>= pS3->pixMuxShift;
+			mode->CrtcHBlankStart >>= pS3->pixMuxShift;
+			mode->CrtcHBlankEnd >>= pS3->pixMuxShift;
 			mode->CrtcHSyncStart >>= pS3->pixMuxShift;
 			mode->CrtcHSyncEnd >>= pS3->pixMuxShift;
 			mode->CrtcHSkew >>= pS3->pixMuxShift;
@@ -1107,6 +1241,8 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 
 			mode->CrtcHTotal <<= -pS3->pixMuxShift;
 			mode->CrtcHDisplay <<= -pS3->pixMuxShift;
+			mode->CrtcHBlankStart <<= -pS3->pixMuxShift;
+			mode->CrtcHBlankEnd <<= -pS3->pixMuxShift;
 			mode->CrtcHSyncStart <<= -pS3->pixMuxShift;
 			mode->CrtcHSyncEnd <<= -pS3->pixMuxShift;
 			mode->CrtcHSkew <<= -pS3->pixMuxShift;
@@ -1160,13 +1296,12 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	}
 
 	/* We need to set this first - S3 *is* broken */
-	outw(vgaCRIndex, (pVga->CRTC[17] << 8) | 17);
+	outw(vgaCRIndex, (pVga->CRTC[17] << 8) | 17); 
 	for(r=0; r<25; r++)
 		outw(vgaCRIndex, (pVga->CRTC[r] << 8) | r);
 
-	for(r=0; r<9; r++) {
+	for(r=0; r<9; r++)
 		outw(0x3ce, (pVga->Graphics[r] << 8) | r);
-	}
 
 	inb(vgaIOBase + 0x0a);
 
@@ -1193,19 +1328,14 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	if ((pS3->Chipset == PCI_CHIP_964_0) ||
 	    (pS3->Chipset == PCI_CHIP_964_1))
 		new->cr33 = 0x20;
+	else if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX)
+		new->cr33 &= ~0x20;
 	outb(vgaCRReg, new->cr33);
 	
 	new->cr34 = 0x10;
 	outb(vgaCRIndex, 0x34);
 	outb(vgaCRReg, new->cr34);
 
-	if (pS3->SlowDRAMRefresh)
-		new->cr3a = 0xb7;
-	else
-		new->cr3a = 0xb5;
-	outb(vgaCRIndex, 0x3a);
-	outb(vgaCRReg, new->cr3a);
-
 	if (pS3->Chipset != PCI_CHIP_AURORA64VP) {
 		new->cr3b = (pVga->CRTC[0] + pVga->CRTC[4] + 1) / 2;
 		outb(vgaCRIndex, 0x3b);
@@ -1260,7 +1390,7 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 		tmp |= 0x10;
 		break;
 	case 24:
-		tmp |= 0x20;
+		tmp |= 0x20; /* there is no such value in spec s3.txt */
 		break;
 	case 32:
 		tmp |= 0x30;
@@ -1331,6 +1461,16 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	new->cr60 = n;
 	outb(vgaCRReg, new->cr60);
 
+	if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX) {
+		new->cr60 = 255;		
+		outb(vgaCRIndex, 0x60);
+		outb(vgaCRReg, new->cr60);		
+
+		new->cr54 = 31 << 3;
+		outb(vgaCRIndex, 0x54);
+		outb(vgaCRReg, new->cr54);		
+	}
+
 	outb(vgaCRIndex, 0x55);
 	new->cr55 = (inb(vgaCRReg) & 0x08) | 0x40;
 	outb(vgaCRReg, new->cr55);
@@ -1345,10 +1485,10 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	{
 		int i;
 		unsigned int j;
-
-		i = ((((mode->CrtcHTotal >> 3) - 5) & 0x100) >> 8) 	|
-		    ((((mode->CrtcHDisplay >> 3) - 1) & 0x100) >> 7) 	|
-		    ((((mode->CrtcHSyncStart >> 3) - 1) & 0x100) >> 6)	|
+		
+		i = ((((mode->CrtcHTotal >> 3) - 5) & 0x100) >> 8)      |
+		    ((((mode->CrtcHDisplay >> 3) - 1) & 0x100) >> 7)    |
+		    ((((mode->CrtcHSyncStart >> 3) - 1) & 0x100) >> 6)  |
 		    ((mode->CrtcHSyncStart & 0x800) >> 7);
 		if ((mode->CrtcHSyncEnd >> 3) - (mode->CrtcHSyncStart >> 3) > 64)
 			i |= 0x08;
@@ -1410,6 +1550,33 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 		outb(vgaCRReg, new->cr42);
 	}
 
+	if (pS3->Chipset == PCI_CHIP_TRIO64V2_DXGX) {
+		unsigned char a;
+
+		outb(vgaCRIndex, 0x67);
+		a = inb(vgaCRReg) & 0xfe;
+
+		switch (pScrn->depth) {
+		case 8:
+			break;
+		case 15:
+			a |= (3 << 4);
+			break;
+		case 16:
+			a |= (5 << 4);
+			break;
+		case 24:
+			a |= (13 << 4);
+			break;		 
+		}
+
+		if (pS3->hasStreams)
+			a |= (3 << 2);
+
+		WaitVSync(); /* Wait for VSync before setting mode */
+		outb(vgaCRReg, a);
+	}	    
+
 	if (pS3->Chipset == PCI_CHIP_968) {
 		unsigned char a;
 
@@ -1471,6 +1638,35 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 		new->cr66 |= 0x80;
 	outb(vgaCRReg, new->cr66);
 
+	if (pS3->SlowDRAMRefresh)
+		new->cr3a = 0xb7;
+	else
+		new->cr3a = 0xb5;
+	outb(vgaCRIndex, 0x3a);
+	outb(vgaCRReg, new->cr3a);
+
+	/* 
+	   Set 3.5 MCLKs for -RAS low, 2.5 MCLKs for -RAS precharge, 
+	   disable -CAS/-OE adjustment. It seems that cr68 has different 
+	   format for 96x and TRIOs
+	*/
+	if (!((pS3->Chipset == PCI_CHIP_968) || 
+	      (pS3->Chipset == PCI_CHIP_964_0) ||
+	      (pS3->Chipset == PCI_CHIP_964_1))) {
+
+		outb(vgaCRIndex, 0x39);
+		outb(vgaCRReg, 0xa5);
+
+		outb(vgaCRIndex, 0x68);
+		tmp = inb(vgaCRReg) & ~0x0f;
+		outb(vgaCRReg, tmp | 0x0f);
+
+		/* Enable 1-cycle EDO access */
+		outb(vgaCRIndex, 0x36);
+		tmp = inb(vgaCRReg);
+		outb(vgaCRReg, tmp & 0xf3);
+	}
+	
 	if (pS3->SlowVRAM) {
 		/*
 		 * some Diamond Stealth 64 VRAM cards have a problem with
@@ -1488,7 +1684,7 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	if (pS3->SlowDRAM) {
 		/*
 		 * fixes some pixel errors for a SPEA Trio64V+ card
-		 * increas -RAS precharge timing from 2.5 MCLKs
+		 * increase -RAS precharge timing from 2.5 MCLKs
 		 * to 3.5 MCLKs
 		 */
 		outb(vgaCRIndex, 0x39);
@@ -1501,7 +1697,7 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	if (pS3->SlowEDODRAM) {
 		/*
 		 * fixes some pixel errors for a SPEA Trio64V+ card
-		 * increas from 1-cycle to 2-cycle EDO mode
+		 * increase from 1-cycle to 2-cycle EDO mode
 		 */
 		outb(vgaCRIndex, 0x39);
 
@@ -1584,13 +1780,8 @@ static Bool S3ModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	WaitQueue(5);
 	SET_SCISSORS(0, 0, pS3->s3ScissR, pS3->s3ScissB);
 
-	outb(vgaCRIndex, 0x6f);
-
-#if 0
-	if (((pScrn->bitsPerPixel == 16) || 
-	     (pScrn->bitsPerPixel == 24)) && (pS3->S3NewMMIO))
+	if (pS3->hasStreams)
 		S3InitStreams(pScrn, mode);
-#endif
 
 	return TRUE;
 }
@@ -1713,7 +1904,7 @@ static void S3AdjustFrame(int scrnIndex, int x, int y, int flags)
 	{
 		int px, py, a;
 
-		miPointerPosition(&px, &py);
+		miPointerGetPosition(inputInfo.pointer, &px, &py);
 
 		if (pS3->s3Bpp == 1)
 			a = 4 - 1;
@@ -1742,7 +1933,7 @@ void S3Regdump(ScrnInfoPtr pScrn)
         S3Ptr pS3 = S3PTR(pScrn);
         int vgaCRIndex = pS3->vgaCRIndex, vgaCRReg = pS3->vgaCRReg;
 
-#if 1
+#if 0
 	outb(vgaCRIndex, 0x31);
 	ErrorF("cr31 = 0x%x\n", inb(vgaCRReg));
 	outb(vgaCRIndex, 0x32);
@@ -1848,11 +2039,47 @@ void S3BankZero(ScrnInfoPtr pScrn)
 	outb(vgaCRReg, tmp);
 }
 
-
-
-static void S3DisplayPowerManagementSet(ScrnInfoPtr pScrn,
-					int PowerManagementMode, int flags)
+static void
+S3DisplayPowerManagementSet(ScrnInfoPtr pScrn, int PowerManagementMode,
+			    int flags)
 {
-	vgaHWDPMSSet(pScrn, PowerManagementMode, flags);
-}
+     S3Ptr pS3 = S3PTR(pScrn);
+     switch (pS3->Chipset) {
+     case PCI_CHIP_TRIO64V2_DXGX:
+     case PCI_CHIP_TRIO:
+     case PCI_CHIP_AURORA64VP:
+     case PCI_CHIP_TRIO64UVP:
+     {
+	  int srd;
+      
+	  outb(0x3c4, 0x08);
+	  outb(0x3c5, 0x06);	  /* unlock extended sequence registers */
 
+	  outb(0x3c4, 0x0d);
+	  srd = inb(0x3c5) & 0xf;  /* clear the sync control bits */
+  
+	  switch (PowerManagementMode) {
+	  case DPMSModeOn:
+	       /* Screen: On; HSync: On, VSync: On */
+	       break;
+	  case DPMSModeStandby:
+	       /* Screen: Off; HSync: Off, VSync: On */
+	       srd |= 0x10;
+	       break;
+	  case DPMSModeSuspend:
+	       /* Screen: Off; HSync: On, VSync: Off */
+	       srd |= 0x40;
+	       break;
+	  case DPMSModeOff:
+	       /* Screen: Off; HSync: Off, VSync: Off */
+	       srd |= 0x50;
+	       break;
+	  }
+	  outb(0x3c4, 0x0d);
+	  outb(0x3c5, srd);
+	  break;
+     }
+     default:
+	  vgaHWDPMSSet(pScrn, PowerManagementMode, flags);
+     }
+}
